import re

def read_chunk(iterator, chunksize=64):
    chunk = []
    for x in range(chunksize):
        try:
            chunk.append(next(iterator))
        except StopIteration:
            return chunk

    return chunk


def get_element(data_dict: dict, path: str, split_element: str = "."):
    if path is None:
        return data_dict

    if callable(path):
        elem = path(data_dict)

    if isinstance(path, str):
        elem = data_dict
        try:
            for x in path.strip(split_element).split(split_element):
                try:
                    x = int(x)
                    elem = elem[x]
                except ValueError:
                    elem = elem.get(x)
        except:
            pass

    if isinstance(path, (list, set)):
        elem = [get_element(data_dict, x) for x in path]

    return elem


def convert_name(name):
    s1 = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", s1).lower()



import importlib
import os
import re
import sys
import logging
import uuid

# from typing import Union

from .plugin import Plugin

from packaging import version


class Manager:
    def __init__(self, configs=None):
        self.configs = configs
        if configs is None:
            self.configs = []

    def plugins(self):
        return {}

    def init_plugins(self, plugins=None, configs=None):
        if plugins is None:
            plugins = list(self.plugins().keys())

        # TODO add merge tools
        if configs is None:
            configs = self.configs

        plugin_list = []
        plugin_name_list = [x.lower() for x in plugins]

        for plugin_name, plugin_class in self.plugins().items():
            if plugin_name.lower() not in plugin_name_list:
                continue
            plugin_has_config = False
            plugin_config = {"params": {}}
            for x in configs:
                if x["plugin"].lower() == plugin_name.lower():
                    plugin_config.update(x)
                    plugin_has_config = True
            if not plugin_has_config:
                continue
            plugin = plugin_class(config=plugin_config["params"])
            plugin_list.append({"plugin": plugin, "plugin_cls": plugin_class, "config": plugin_config})
        return plugin_list



import imageio.v3 as iio
import av
import numpy as np


def parse_meta_av(path, **kwargs):
    try:
        fh = av.open(path)
        stream = fh.streams.video[0]
        frame = next(fh.decode(video=0))
        frame = frame.reformat(format="rgb24")
        return {
            "fps": stream.average_rate,
            "width": frame.width,
            "height": frame.height,
            "size": (frame.width, frame.height),
            "duration": float(stream.duration * stream.time_base),
        }

    except:
        return None


def parse_meta_imageio(path, **kwargs):
    try:
        # TODO
        meta = iio.immeta(path, plugin="FFMPEG", **kwargs)
        return {
            "fps": meta.average_rate,
            "width": meta.width,
            "height": meta.height,
            "size": (meta.width, meta.height),
            "duration": float(meta.duration * meta.time_base),
        }

    except:
        return None


class VideoDecoder:
    # TODO: videos with sample aspect ratio (SAR) not equal to 1:1 are loaded with wrong shape
    def __init__(self, path, max_dimension=None, fps=None, ref_id=None, **kwargs):
        """Provides an iterator over the frames of a video.

        Args:
            path (str): Path to the video file.
            max_dimension (Union[List, int], optional): Resize the video by providing either
                - a list of shape [width, height] or
                - an int that depicts the longer side of the frame.
                Defaults to None.
            fps (int, optional): Frames per second. Defaults to None.
        """
        self._path = path
        self._max_dimension = max_dimension
        self._fps = fps
        self._ref_id = ref_id

        self._meta = parse_meta_av(path)

        self._size = self._meta.get("size")

        self._real_fps = self._meta.get("fps")
        self._duration = self._meta.get("duration")

        self._kwargs = kwargs

    def __iter__(self):
        filter_sequence = []

        if self._fps is not None:
            filter_sequence.append(("fps", {"fps": f"{self._fps}", "round": "up"}))

        if self._max_dimension is not None:
            if isinstance(self._max_dimension, (list, tuple)):
                res = self._max_dimension
            else:
                res = max(self._size[1], self._size[0])

                scale = min(self._max_dimension / res, 1)
                res = (round(self._size[0] * scale), round(self._size[1] * scale))
            filter_sequence.append(
                ("scale", {"width": f"{res[0]}", "height": f"{res[1]}"})
            )
        video_reader = iio.imiter(
            self._path,
            plugin="pyav",
            format="rgb24",
            filter_sequence=filter_sequence,
            **self._kwargs,
        )
        fps = self._real_fps if self._fps is None else self._fps
        for i, frame in enumerate(video_reader):
            yield {
                "time": float(i / fps),
                "index": i,
                "frame": frame,
                "ref_id": self._ref_id,
                "delta_time": float(1 / fps),
            }

    def __len__(self):
        return self.duration() * self.fps()

    def fps(self):
        return float(self._real_fps if self._fps is None else self._fps)

    def duration(self):
        return self._duration


class VideoBatcher:
    def __init__(self, video_decoder: VideoDecoder, batch_size=8):
        self.video_decoder = video_decoder
        self.batch_size = batch_size

    def __iter__(self):
        cache = []
        for x in self.video_decoder:
            cache.append(x)

            if len(cache) >= self.batch_size:
                yield {
                    "time": [x["time"] for x in cache],
                    "index": [x["index"] for x in cache],
                    "frame": np.stack([x["frame"] for x in cache]),
                    "ref_id": [x["ref_id"] for x in cache],
                }
                cache = []

        if len(cache) >= self.batch_size:
            yield {
                "time": [x["time"] for x in cache],
                "index": [x["index"] for x in cache],
                "frame": np.stack([x["frame"] for x in cache]),
                "ref_id": [x["ref_id"] for x in cache],
            }

    def fps(self):
        return self.video_decoder._fps

    def duration(self):
        return self.video_decoder._duration



from packaging import version
from typing import Union, Dict, Any

from utils.helper import convert_name


class Plugin:
    @classmethod
    def __init_subclass__(
        cls,
        config: Dict[str, Any] = None,
        parameters: Dict[str, Any] = None,
        version: str = None,
        **kwargs,
    ):
        super().__init_subclass__(**kwargs)
        cls._default_config = config
        cls._version = version
        cls._parameters = parameters
        cls._name = convert_name(cls.__name__)

    def __init__(self, config=None):
        self._config = self._default_config
        if config is not None:
            self._config.update(config)

    @property
    def config(self):
        return self._config

    @classmethod
    @property
    def default_config(cls):
        return cls._default_config

    @classmethod
    @property
    def version(cls):
        return cls._version

    @classmethod
    @property
    def name(cls):
        return cls._name



import imageio
import numpy as np
from PIL import Image


def image_normalize(image):
    if len(image.shape) == 2:

        return np.stack([image] * 3, -1)

    if len(image.shape) == 3:
        if image.shape[-1] == 4:
            return image[..., 0:3]
        if image.shape[-1] == 1:
            return np.concatenate([image] * 3, -1)

    if len(image.shape) == 4:
        return image_normalize(image[0, ...])

    return image


def image_crop(image, size=None):

    image = Image.fromarray(image)
    width, height = image.size  # Get dimensions

    left = (width - size[0]) / 2
    top = (height - size[1]) / 2
    right = (width + size[0]) / 2
    bottom = (height + size[1]) / 2

    # Crop the center of the image
    image = image.crop((left, top, right, bottom))
    return np.array(image)


def image_pad(image):
    height, width = image.shape[:2]  # WHC

    pad_x = max(0, (height - width) // 2)
    pad_y = max(0, (width - height) // 2)

    return np.pad(image, ((pad_y, pad_y), (pad_x, pad_x), (0, 0)), "constant", constant_values=0)


def image_from_proto(image_proto):
    image_field = image_proto.WhichOneof("image")
    if image_field == "path":
        image = imageio.imread(image_proto.path)
    if image_field == "encoded":
        image = imageio.imread(image_proto.encoded)

    image = image_normalize(image)

    if image_proto.roi.width > 0 and image_proto.roi.height > 0:
        y = int(image_proto.roi.y * image.shape[0])
        height = int(image_proto.roi.height * image.shape[0])
        x = int(image_proto.roi.x * image.shape[1])
        width = int(image_proto.roi.width * image.shape[1])

        image = image[y : y + height, x : x + width]

    return image


def image_resize(image, max_dim=None, min_dim=None, size=None):
    if max_dim is not None:
        shape = np.asarray(image.shape[:2], dtype=np.float32)

        long_dim = max(shape)
        scale = min(1, max_dim / long_dim)
        new_shape = np.asarray(shape * scale, dtype=np.int32)

    elif min_dim is not None:
        shape = np.asarray(image.shape[:2], dtype=np.float32)

        short_dim = min(shape)
        scale = max(1, min_dim / short_dim)
        new_shape = np.asarray(shape * scale, dtype=np.int32)
    elif size is not None:
        new_shape = size
    else:
        return image

    Image.MAX_IMAGE_PIXELS = 10000 * 10000
    img = Image.fromarray(image)
    img = img.resize(size=new_shape[::-1])
    return np.array(img)



import logging
from typing import Dict

from utils.plugin import Plugin
from utils.plugin import Factory


class Cache(Plugin):
    def __init__(self, config=None):
        super().__init__(config)


class CacheManager(Factory):
    _plugins = {}

    @classmethod
    def export(cls, name: str):
        def export_helper(plugin):
            cls._plugins[name] = plugin
            return plugin

        return export_helper

    @classmethod
    def build(cls, name: str, config: Dict = None):
        if name not in cls._plugins:
            logging.error(f"Unknown cache server: {name}")
            return None

        return cls._plugins[name](config)


import hashlib
import json
import logging
from typing import List, Dict


def flat_dict(data_dict, parse_json=False):
    result_map = {}
    for k, v in data_dict.items():
        if isinstance(v, dict):
            embedded = flat_dict(v)
            for s_k, s_v in embedded.items():
                s_k = f"{k}.{s_k}"
                if s_k in result_map:
                    logging.error(f"flat_dict: {s_k} alread exist in output dict")

                result_map[s_k] = s_v
            continue

        if k not in result_map:
            result_map[k] = []
        result_map[k] = v
    return result_map


def get_hash_for_plugin(
    plugin: str,
    output: str,
    version: str = None,
    parameters: List = [],
    inputs: List = [],
    config: Dict = {},
):
    plugin_call_dict = {
        "plugin": plugin,
        "output": output,
        "parameters": parameters,
        "inputs": inputs,
        "config": config,
        "version": version,
    }
    # logging.info(f"[HASH] {plugin_call_dict}")

    # logging.info(f"[HASH] {flat_dict(plugin_call_dict)}")
    plugin_hash = hashlib.sha256(
        json.dumps(
            flat_dict(
                {
                    "plugin": plugin,
                    "output": output,
                    "parameters": parameters,
                    "inputs": inputs,
                    "config": config,
                    "version": version,
                }
            )
        ).encode()
    ).hexdigest()

    # logging.info(f"[HASH] {plugin_hash}")
    return plugin_hash



class ByteFIFO:
    def __init__(self):
        self.buffer = bytearray()

    def write(self, data: bytes) -> None:
        self.buffer.extend(data)

    def read(self, chunk_size: int = None) -> bytes:
        if not chunk_size:
            data = self.buffer[:]
            del self.buffer[:]
            return bytes(data)
        data = self.buffer[:chunk_size]
        del self.buffer[:chunk_size]
        return bytes(data)

    def __len__(self):
        return len(self.buffer)



import logging
from typing import Dict


class Factory:
    _plugins = {}

    @classmethod
    def export(cls, name: str):
        def export_helper(plugin):
            cls._plugins[name] = plugin
            return plugin

        return export_helper

    @classmethod
    def build(cls, name: str, config: Dict = None):
        if name not in cls._plugins:
            logging.error(f"Unknown inference server: {name}")
            return None

        return cls._plugins[name](config)



from typing import Any, List, Iterator
import logging

import valkey
import msgpack

from utils.cache import CacheManager, Cache

default_config = {"db": 0, "host": "valkey", "port": 6379, "tag": "data"}


class Batcher:
    def __init__(self, iterable, n=1):
        self.iterable = iterable
        self.n = n

    def __iter__(self):
        l = len(self.iterable)
        for ndx in range(0, l, self.n):
            yield self.iterable[ndx : min(ndx + self.n, l)]


@CacheManager.export("valkey")
class valkeyCache(Cache, config=default_config, version="0.1"):
    def __init__(self, config=None):
        super().__init__(config)
        self.r = valkey.Valkey(
            host=self.config.get("host"),
            port=self.config.get("port"),
            db=self.config.get("db"),
        )

    def set(self, id: str, data: Any) -> bool:
        try:
            packed = msgpack.packb(data)
            tag = self.config.get("tag")
            self.r.set(f"{tag}:{id}", packed)
        except Exception as e:
            logging.error(f"valkeyCache {e}")

    def delete(self, id: str) -> bool:
        try:
            tag = self.config.get("tag")
            return self.r.delete(f"{tag}:{id}")
        except Exception as e:
            logging.error(f"valkeyCache {e}")
            return None

    def get(self, id: str) -> Any:
        try:
            tag = self.config.get("tag")
            packed = self.r.get(f"{tag}:{id}")
            if packed is None:
                return None
            return msgpack.unpackb(packed)
        except Exception as e:
            logging.error(f"valkeyCache {e}")
            return None

    def keys(self) -> List[str]:
        try:
            tag = self.config.get("tag")
            start = len(f"{tag}:")
            keys = self.r.scan_iter(f"{tag}:*", 500)

            # print([x for x in Batcher(keys, 2)])
            return [key[start:].decode("utf-8") for key in keys]
        except Exception as e:
            logging.error(f"valkeyCache {e}")
            return []

    def __iter__(self) -> Iterator:
        try:
            tag = self.config.get("tag")
            start = len(f"{tag}:")
            keys = list(self.r.scan_iter(f"{tag}:*", 500))
            while len(keys) > 0:
                batch_keys = keys[:500]
                keys = keys[500:]

                values = self.r.mget(batch_keys)
                for k, v in zip(batch_keys, values):
                    yield k[start:].decode("utf-8"), msgpack.unpackb(v)

        except Exception as e:
            logging.error(f"valkeyCache {e}")
            yield from []
